/*
Author Klaus Triendl (klaus.triendl@firedaemon.com)
Copyright FireDaemon Technologies Limited 2015.
Distributed under the Boost Software License, Version 1.0.
(See http://www.boost.org/LICENSE_1_0.txt)
*/

#pragma once
#include <atomic>
#include <memory>
#include <functional>
#include <assert.h>


typedef std::unique_ptr<void, int(__stdcall*)(void*)> handle_ptr;


/** @short Watches a given handle once until it becomes signaled, whereupon the callback (work item) gets called.
 */
class handle_watcher
{
public:
    handle_watcher();
    ~handle_watcher()
    {
        if (handle_ptr handle = cleanup())
            // not perfect: cleanup() should have been called on the thread that installed the watch;
            // explicit cleanup is much more obvious and safe in regards to thread-safety
            assert(!handle);
    }

    /** @short Take ownership of the handle and register it to be waited on in the system thread pool.
        The callback (work item) is executed once the handle becomes signaled;
        it is reset as soon as it has been executed. If you have special bound parameters, ensure
        that destroying them is thread-safe.
        @return Whether the passed in handle is valid
    */
    bool watch_this(handle_ptr handle, std::function<void()> workItem);
    bool watch_this(void*&& handle, std::function<void()> workItem);

    /** @short unregisters conditionally (waiting for already queued callbacks to get called)
        @return The returned handle comes in handy when you still need to use it for post-flight operations.
        Note that the returned handle is only valid when the callback wasn't called before, otherwise it is null
     */
    handle_ptr cleanup();

private:
    // internal callback handler
    static void __stdcall dispatch_signaled(void* context, unsigned char timedOut)
    {
        handle_watcher* const _this = static_cast<handle_watcher*>(context);
        _this->signaled();
    }

    void signaled();

private:
    std::atomic<void*> waitObject_;
    handle_ptr handle_;
    std::function<void()> workItem_;
};
